/*
 * libxml-dtc provides an XML framework creation facilities
 * Copyright (C) 2003-2004 Mike Hommey <mh@glandium.org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#define _GNU_SOURCE

#include <string.h>
#include <stdlib.h>
#include <mappings.h>

#define SIZE_MOD 64

int xml_dtc_mappings_free(xml_dtc_mappings *mappings) {
  if (mappings == NULL)
    return -1;

  if (mappings->resource)
    xml_dtc_resource_unlink(mappings->resource);
  if (mappings->list && mappings->type) {
    int i;
    for (i = 0; i < mappings->len; i++) {
      if (mappings->type[i] == m_hash) {
	xml_dtc_hash_free((xml_dtc_hash *)mappings->list[i]);
	mappings->list[i] = NULL;
      } else
      if (mappings->type[i] == m_list) {
	xml_dtc_list_free((xml_dtc_list *)mappings->list[i]);
	mappings->list[i] = NULL;
      }
    }
    free(mappings->list);
    free(mappings->type);
  }

  free(mappings);
  return 0;
}

xml_dtc_mappings *xml_dtc_mappings_new(const xml_dtc_resource *resource) {
  xml_dtc_mappings *result;

  if (resource == NULL)
    return NULL;
  
  result = (xml_dtc_mappings *) malloc(sizeof(xml_dtc_mappings));
  memset(result, 0, sizeof(xml_dtc_mappings));
  result->resource = xml_dtc_resource_link(resource);

  return result;
}

char *xml_dtc_resolve_mappings(xml_dtc_mappings *mappings, char *string) {
  int i, j, begin, end; char result[4096];

  if (mappings && string) {
    for (i = 0, j = 0; (string[i] != 0) && (j < 4095); i++) {
      if (string[i] == '{') {
	begin = i;
	while (string[++i] != '}' && string[i] != 0) { }
	if (string[i] != 0) {
	  char *val, *tmp;
	  int offset, k = 0, is_int = 1;
	  end = i;
	  val = strndup(&string[begin + 1], end - begin - 1);
	  tmp = val;
	  if (mappings->len > 0) {
	    offset = mappings->len - 1;
	    while(strncmp(tmp, "../", 3) == 0) {
	      tmp += 3;
	      offset--;
	    }
	    if (offset < 0)
	      offset = 0;
	    while(tmp[k] != 0) {
	      if ((tmp[k] < '0') || (tmp[k] > '9')) {
		is_int = 0;
		break;
	      }
	      k++;
	    }
	    if (is_int) {
	      begin = atoi(tmp);
	      if (mappings->type[offset] == m_list)
		tmp = (char *)xml_dtc_list_get((xml_dtc_list *)mappings->list[offset], begin);
	      else
		tmp = NULL;
	    } else {
	      if (mappings->type[offset] == m_hash)
		tmp = (char *)xml_dtc_hash_get((xml_dtc_hash *)mappings->list[offset], tmp);
	      else
		tmp = NULL;
	    }
	    free(val);
	    if (tmp) {
	      memcpy(&result[j], tmp, strlen(tmp));
	      j += strlen(tmp);
	      xml_dtc_msg(mappings->resource, XML_DTC_MSG_DEBUG, "Using %s", val);
	    } else {
	      xml_dtc_msg(mappings->resource, XML_DTC_MSG_DEBUG, "No such mapping");
	    }
	  } else {
	    xml_dtc_msg(mappings->resource, XML_DTC_MSG_DEBUG, "No such mapping");
	  }
	}
      } else
	result[j++] = string[i];
    }
    result[j++] = 0;
    
    xml_dtc_msg(mappings->resource, XML_DTC_MSG_DEBUG, "Resolved '%s' to '%s'", string, result);
    return strdup(result);
  } else {
    return NULL;
  }
}

static int xml_dtc_mappings_push(xml_dtc_mappings *mappings, void *element, xml_dtc_mappings_type type) {
  if (mappings && element) {
    if ((mappings->len % SIZE_MOD) == 0) {
      mappings->list = realloc(mappings->list, sizeof(void *) * (mappings->len + SIZE_MOD));
      mappings->type = realloc(mappings->type, sizeof(xml_dtc_mappings_type) * (mappings->len + SIZE_MOD));
    }
    mappings->list[mappings->len] = element;
    mappings->type[mappings->len++] = type;
    return 0;
  } else {
    return -1;
  }
}

xml_dtc_list *xml_dtc_mappings_new_list(xml_dtc_mappings *mappings) {
  xml_dtc_list *result;

  result = xml_dtc_string_list_new(mappings->resource);
  xml_dtc_mappings_push(mappings, result, m_list);
  return result;
}

xml_dtc_hash *xml_dtc_mappings_new_hash(xml_dtc_mappings *mappings) {
  xml_dtc_hash *result;

  result = xml_dtc_string_hash_new(mappings->resource);
  xml_dtc_mappings_push(mappings, result, m_hash);
  return result;
}

int xml_dtc_mappings_destroy_last(xml_dtc_mappings *mappings) {
  if (mappings && mappings->len) {
    mappings->len--;
    if (mappings->type[mappings->len] == m_hash)
      xml_dtc_hash_free(mappings->list[mappings->len]);
    else if (mappings->type[mappings->len] == m_list)
      xml_dtc_list_free(mappings->list[mappings->len]);
    mappings->list[mappings->len] = NULL;
    return 0;
  } else {
    return -1;
  }
}
