/*
 * libxml-dtc provides an XML framework creation facilities
 * Copyright (C) 2003-2004 Mike Hommey <mh@glandium.org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#define _GNU_SOURCE
#include <sitemap.h>
#include <string.h>

int xml_dtc_sitemap_add(xml_dtc_sitemap *sitemap, xml_dtc_component *object) {
  if (sitemap && object) {
    xml_dtc_hash *container = NULL;
    container = (xml_dtc_hash *)xml_dtc_hash_get(sitemap->hash, object->type);
    if (container == NULL) {
      container = xml_dtc_component_hash_new(sitemap->resource);
      xml_dtc_hash_add_cast(sitemap->hash, object->type, container);
    }
    xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Adding '%s' to sitemap", object->type);
    return xml_dtc_component_hash_add(container, object);
  } else {
    return -1;
  }
}

xml_dtc_component *xml_dtc_sitemap_get(xml_dtc_sitemap *sitemap, char *type, char *name) {
  return xml_dtc_hash_get(xml_dtc_sitemap_get_all_type(sitemap, type), name);
}

xml_dtc_hash *xml_dtc_sitemap_get_all_type(xml_dtc_sitemap *sitemap, char *type) {
  if (sitemap)
    if (type)
      return (xml_dtc_hash *)xml_dtc_hash_get(sitemap->hash, type);
  return NULL;
}

int xml_dtc_sitemap_set_preferred_by_key(xml_dtc_sitemap *sitemap, char *type, char *name) {
  if (sitemap && type && name)
    return xml_dtc_hash_set_default((xml_dtc_hash *)xml_dtc_hash_get(sitemap->hash, type), name);
  return -1;
}

int xml_dtc_sitemap_free(xml_dtc_sitemap *sitemap) {
  if (sitemap == NULL)
    return -1;

  if (sitemap->resource)
    xml_dtc_resource_unlink(sitemap->resource);

  if (sitemap->base_dir)
    free(sitemap->base_dir);

  if (sitemap->hash)
    xml_dtc_hash_free(sitemap->hash);

  if (sitemap->pipelines)
    xml_dtc_list_free(sitemap->pipelines);

  free(sitemap);

  return 0;
}

xml_dtc_sitemap *xml_dtc_sitemap_new(const xml_dtc_resource *resource) {
  xml_dtc_sitemap *result;

  if (resource == NULL)
    return NULL;

  result = (xml_dtc_sitemap *)malloc(sizeof(xml_dtc_sitemap));
  memset(result, 0, sizeof(xml_dtc_sitemap));
  result->resource = xml_dtc_resource_link(resource);

  result->hash = xml_dtc_hash_new_cast(resource, free, strdup, xml_dtc_hash_free, NULL);

  return result;
}

xml_dtc_list *xml_dtc_sitemap_get_pipeline(xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xml_dtc_caller_def *caller;
  xml_dtc_list *result = xml_dtc_list_new_cast(sitemap->resource, xml_dtc_caller_unlink, NULL);

  while (node) {
    /* Ignore Blank Nodes */
    if (! xmlIsBlankNode(node)) {
      /* Test if node is an element or something else (PI, Comment...) */
      if (node->type != XML_ELEMENT_NODE) {
        /* Comments are allowed */
        if (node->type == XML_COMMENT_NODE) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Found comment");
        } else {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Element or comment expected");
          xml_dtc_list_free(result);
          return NULL;
        }
      } else {
        /* Check if namespace is good */
        if ((node->ns != node->parent->ns) && (xmlStrcmp(node->ns->href, XML_DTC_SITEMAP_NAMESPACE) != 0)) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "expecting %s namespace, but got %s", XML_DTC_SITEMAP_NAMESPACE,
                      (node->ns != NULL) ? (char *) node->ns->href : "none");
          xml_dtc_list_free(result);
          return NULL;
        }
        /* Now we know that namespace is ok, we can check and initialize all stuff */
        caller = xml_dtc_hash_get(xml_dtc.callers, node->name);
        if (! caller) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "Unexpected node: %s", node->name);
          xml_dtc_list_free(result);
          return NULL;
        }
        xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Node %s -> %s.", node->name, caller->component);
        if (caller->callback) {
          xml_dtc_list_push(result, caller->callback(NULL, NULL, sitemap, node));
        }
      }
    }
    node = node->next;
  }

  return result;
}

static int xml_dtc_sitemap_loop(xml_dtc_sitemap *sitemap, xmlNodePtr node, const int depth) {
  while (node) {
    /* Ignore Blank Nodes */
    if (! xmlIsBlankNode(node)) {
      /* Test if node is an element or something else (PI, Comment...) */
      if (node->type != XML_ELEMENT_NODE) {
        /* Comments are allowed */
        if (node->type == XML_COMMENT_NODE) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Found comment");
        } else {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Element or comment expected");
          return -1;
        }
      } else {
        /* Check if namespace is good */
        if ((node->ns != node->parent->ns) && (xmlStrcmp(node->ns->href, XML_DTC_SITEMAP_NAMESPACE) != 0)) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "expecting %s namespace, but got %s", XML_DTC_SITEMAP_NAMESPACE,
                      (node->ns != NULL) ? (char *) node->ns->href : "none");
          return -1;
        }
        /* Now we know that namespace is ok, we can check and initialize all stuff */
        xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Node %s.", node->name);

        if ((depth == 0) && (xmlStrcmp(node->name, "components") == 0)) {
          if (xml_dtc_sitemap_loop(sitemap, node->children, depth + 1) != 0)
            return -1;
        } else if ((depth == 0) && (xmlStrcmp(node->name, "pipelines") == 0)) {
          if (xml_dtc_sitemap_loop(sitemap, node->children, depth + 1) != 0)
            return -1;
        } else if ((depth == 0) || ((depth == 1) && (xmlStrcmp(node->parent->name, "components") == 0))) {
          xmlChar *name;
          char *type;
          name = xmlStrndup(node->name, xmlStrlen(node->name) - 1);
          type = xml_dtc_hash_get_key(xml_dtc.components, name);
          xmlFree(name);
          if (type) {
            xmlChar *deflt;
            if (xml_dtc_sitemap_loop(sitemap, node->children, depth + 1))
              return -1;
            deflt = xmlGetProp(node, "default");
            if (deflt) {
              xml_dtc_sitemap_set_preferred_by_key(sitemap, type, deflt);
              xmlFree(deflt);
            }
          } else
            goto broken;
        } else if ((depth == 1) && (xmlStrcmp(node->parent->name, "pipelines") == 0) && (xmlStrcmp(node->name, "pipeline") == 0)) {
          if (!sitemap->pipelines)
            sitemap->pipelines = xml_dtc_list_new_cast(sitemap->resource, xml_dtc_list_free, NULL);
          xml_dtc_list_push(sitemap->pipelines, xml_dtc_sitemap_get_pipeline(sitemap, node->children));
        } else if ((depth == 1) && (xmlStrncmp(node->parent->name, node->name, xmlStrlen(node->name)) == 0)  && (xmlStrlen(node->name) == xmlStrlen(node->parent->name) -1)) {
          xml_dtc_component_def *component;
          xml_dtc_hash *hash = xml_dtc_hash_get(xml_dtc.components, node->name);
          xmlChar *name;

          name = xmlGetProp(node, "name");
          if (!name) {
            xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "Missing attribute: name");
            return -1;
          }
          component = xml_dtc_hash_get(hash, node->name);
          if (!component) {
            xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "Unknown type: %s", node->name);
            xmlFree(name);
            return -1;
          }
          if (component->callback) {
            /* BOUH QUE C'EST PA BO */
            component->callback((xml_dtc_handler *)&component->handle, NULL, sitemap, node);
          }
          xmlFree(name);
        } else if ((depth == 2) && (xmlStrcmp(node->parent->parent->name, "components") == 0) && (xmlStrncmp(node->parent->name, node->name, xmlStrlen(node->name)) == 0) && (xmlStrlen(node->name) == xmlStrlen(node->parent->name) - 1)) {
          xml_dtc_component_def *component;
          xml_dtc_hash *hash = xml_dtc_hash_get(xml_dtc.components, node->name);
          xmlChar *src;
          xmlChar *name;
          if (!hash)
            goto broken;
          src = xmlGetProp(node, "src");
          if (!src) {
            xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "Missing attribute: src");
            return -1;
          }
          name = xmlGetProp(node, "name");
          if (!name) {
            xmlFree(src);
            xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "Missing attribute: name");
            return -1;
          }
          component = xml_dtc_hash_get(hash, src);
          if (!component) {
            xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "Unknown %s: %s", node->name, src);
            xmlFree(src);
            xmlFree(name);
            return -1;
          }
          if (component->callback) {
            /* BOUH QUE C'EST PA BO */
            component->callback((xml_dtc_handler *)&component->handle, NULL, sitemap, node);
          }
          xmlFree(src);
          xmlFree(name);
        } else {
          goto broken;
        }
      }
    }
    node = node->next;
  }

  return 0;

 broken:
  xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "Unexpected node: %s", node->name);
  return -1;
}

xml_dtc_sitemap *xml_dtc_sitemap_new_from_file(xml_dtc_resource *resource, char *file) {
  xml_dtc_sitemap *result = NULL;
  xmlDocPtr xmldoc;
  xmlNodePtr node;
  char *tmp;

  xml_dtc_msg(resource, XML_DTC_MSG_DEBUG, "Opening file '%s'.", file);

  xmldoc = xmlParseFile(file);

  if (xmldoc == NULL) {
    xml_dtc_msg(resource, XML_DTC_MSG_ERROR, "File '%s' not parsed successfully.", file);
    return NULL;
  }

  node = xmlDocGetRootElement(xmldoc);
  if (node == NULL) {
    xml_dtc_msg(resource, XML_DTC_MSG_ERROR, "File '%s' does not have a root element.", file);
    xmlFreeDoc(xmldoc);
    return NULL;
  }

  if (xmlStrcmp(node->ns->href, XML_DTC_SITEMAP_NAMESPACE)
      || xmlStrcmp(node->name, (const xmlChar *) XML_DTC_SITEMAP_ROOT_NODE)) {
    xml_dtc_msg(resource, XML_DTC_MSG_ERROR, "File doesn't have the right root node.");
    xmlFreeDoc(xmldoc);
    return NULL;
  }
  
  result = xml_dtc_sitemap_new(resource);

  xml_dtc_sitemap_loop(result, node->children, 0);
  
  xmlFreeDoc(xmldoc);

  if ((tmp = rindex(file, '/'))) {
    result->base_dir = strndup(file, (int)tmp - (int)file);
  }
  return result;
}

int xml_dtc_sitemap_request(xml_dtc_sitemap *sitemap, xml_dtc_request *request) {
  int i;

  if (!xml_dtc.resource)
    return -1;
  if (sitemap && request && sitemap->pipelines) {
    request->sitemap = sitemap;
    for (i = 0; (request->error.type == NULL) && (request->result == NULL) && (i < sitemap->pipelines->len); i++) {
      xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Running pipeline #%d", i);
      xml_dtc_request_apply(request, (xml_dtc_list *)sitemap->pipelines->list[i]);
    }

    if ((request->result == NULL) && (request->error.type == 0)) {
      xml_dtc_request_set_error(request, (xml_dtc_component *)request->sitemap, "resource-not-found", "Resource not found", "No pipeline matched the requested URI.");
    }
    if (request->error.type && ! request->result) {
      return xml_dtc_request_serialize_error(request);
    }
    return 0;
  }
  return -1;
}

int xml_dtc_sitemap_resolve_relative_path(xml_dtc_sitemap *sitemap, char **path) {
  if (sitemap && path && *path) {
    if ((*path[0] != '/') && sitemap->base_dir) {
      char *tmp;
      tmp = malloc(strlen(*path) + strlen(sitemap->base_dir) + 2);
      sprintf(tmp, "%s/%s", sitemap->base_dir, *path);
      free(*path);
      *path = tmp;
      return 0;
    }
    return 0;
  }
  return -1;
}
