/*
 * This module provides matching to the XML-DTC framework
 * Copyright (C) 2003-2004 Mike Hommey <mh@glandium.org>
 * 
 * This module is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this module; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#include <modules.h>
#include <xml-dtc.h>
#include <mappings.h>
#include <list.h>
#include <string.h>
#include <pcre.h>

typedef struct _xml_dtc_matcher_caller_data xml_dtc_matcher_caller_data;

struct _xml_dtc_matcher_caller_data {
  xmlChar *pattern;
  xml_dtc_list *sub_pipeline;
  pcre *re;
};

static int _xml_dtc_matcher_caller_free(xml_dtc_matcher_caller_data *data) {
  if (data) {
    if (data->sub_pipeline)
      xml_dtc_list_free(data->sub_pipeline);
    if (data->pattern)
      xmlFree(data->pattern);
    if (data->re)
      pcre_free(data->re);
    free(data);
  }
  return 0;
}

static xml_dtc_component *xml_dtc_handler_matcher_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xmlChar *attr;
  xml_dtc_component *result;

  attr = xmlGetProp(node, "name");
  result = xml_dtc_component_new(sitemap->resource, attr, (char *)node->name);
  result->handle = handler->_private;
  free(attr);

  xml_dtc_sitemap_add(sitemap, result);
  return (xml_dtc_component *)handler;
}

static xml_dtc_component *xml_dtc_handler_matcher_wildcard_header_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xmlChar *attr;
  xml_dtc_component *result;

  attr = xmlGetProp(node, "name");
  result = xml_dtc_component_new(sitemap->resource, attr, (char *)node->name);
  result->handle = handler->_private;
  free(attr);

  node = node->children;

  while (node) {
    /* Ignore Blank Nodes */
    if (! xmlIsBlankNode(node)) {
      /* Test if node is an element or something else (PI, Comment...) */
      if (node->type != XML_ELEMENT_NODE) {
        /* Comments are allowed */
        if (node->type == XML_COMMENT_NODE) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Found comment");
        } else {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Element or comment expected");
          return NULL;
        }
      } else {
        /* There must be NO namespace */
        if (node->ns) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "expecting no namespace, but got %s",
                      node->ns->href);
          return NULL;
        }
        if (xmlStrcmp(node->name, "header-name") == 0) {
          result->_private = xmlNodeGetContent(node);
          result->free_private = (int (*) (void *)) xmlFree;
        } else {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "Unauthorized configuration %s", node->name);
          return NULL;
        }
      }
    }
    node = node->next;
  }

  xml_dtc_sitemap_add(sitemap, result);
  return (xml_dtc_component *)handler;
}

static xml_dtc_component *xml_dtc_handler_matcher_wildcard_request_parameter_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xmlChar *attr;
  xml_dtc_component *result;

  attr = xmlGetProp(node, "name");
  result = xml_dtc_component_new(sitemap->resource, attr, (char *)node->name);
  result->handle = handler->_private;
  free(attr);

  node = node->children;

  while (node) {
    /* Ignore Blank Nodes */
    if (! xmlIsBlankNode(node)) {
      /* Test if node is an element or something else (PI, Comment...) */
      if (node->type != XML_ELEMENT_NODE) {
        /* Comments are allowed */
        if (node->type == XML_COMMENT_NODE) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Found comment");
        } else {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Element or comment expected");
          return NULL;
        }
      } else {
        /* There must be NO namespace */
        if (node->ns) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "expecting no namespace, but got %s",
                      node->ns->href);
          return NULL;
        }
        if (xmlStrcmp(node->name, "parameter-name") == 0) {
          result->_private = xmlNodeGetContent(node);
	  result->free_private = (int (*) (void *)) xmlFree;
        } else {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "Unauthorized configuration %s", node->name);
          return NULL;
        }
      }
    }
    node = node->next;
  }

  xml_dtc_sitemap_add(sitemap, result);
  return (xml_dtc_component *)handler;
}

static xml_dtc_list *xml_dtc_handler_pipeline_loop_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xml_dtc_list *result;

  result = xml_dtc_sitemap_get_pipeline(sitemap, node->children);

  return result;
}

static xml_dtc_component *xml_dtc_handler_match_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xml_dtc_caller *result;
  xml_dtc_matcher_caller_data *data;

  result = xml_dtc_caller_new(sitemap->resource, "matcher", "matcher");
  data = malloc(sizeof(xml_dtc_matcher_caller_data));
  memset(data, 0, sizeof(xml_dtc_matcher_caller_data));
  result->_private = data;
  result->free_private = (int (*) (void *))_xml_dtc_matcher_caller_free;
  result->component.name = xmlGetProp(node, "type");
  data->pattern = xmlGetProp(node, "pattern");

  /* DIRTY HACK */
  xmlNodeSetName(node, "pipeline");
  data->sub_pipeline = xml_dtc_handler_pipeline_loop_callback(handler, handlers, sitemap, node);

  return (xml_dtc_component *)result;
}

static pcre *pattern2re(char *pattern) {
  int i, j, single = 0, dble = 0, esc = 0;
  const char *error;
  int erroffset;
  char *tmp;
  pcre *re;
  /* Note that there WILL be a problem if original pattern contains parenthesis */
  for (i = 0; pattern[i] != 0; i++) {
    if (pattern[i] == '*') {
      if (pattern[i + 1] == '*') {
	dble++;
	i++;
      } else {
	single++;
      }
    } else if ((pattern[i] == '+') || (pattern[i] == '.')) {
      esc++;
    }
  }
  tmp = malloc(i + single * 6 + dble * 3 + esc + 4);
  tmp[0] = '^';
  for (i = 0, j = 1; pattern[i] != 0; i++) {
    if (pattern[i] == '*')
      if (pattern[i + 1] == '*') {
	i++;
	tmp[j++] = '(';
	tmp[j++] = '.';
	tmp[j++] = '*';
	tmp[j++] = ')';
      } else {
	tmp[j++] = '(';
	tmp[j++] = '[';
	tmp[j++] = '^';
	tmp[j++] = '/';
	tmp[j++] = ']';
	tmp[j++] = '*';
	tmp[j++] = ')';
      }
    else if ((pattern[i] == '+') || (pattern[i] == '.')) {
      tmp[j++] = '\\';
      tmp[j++] = pattern[i];
    }
    else
      tmp[j++] = pattern[i];
  }
  tmp[j++] = '$';
  tmp[j] = 0;
  re = pcre_compile(tmp, PCRE_UTF8, &error, &erroffset, NULL);
  free(tmp);
  return re;
}

static int xml_dtc_match(xml_dtc_caller *caller, xml_dtc_request *request) {
  int rc;
  int ovector[30];
  xml_dtc_matcher_caller_data *data = caller->_private;

  if (!data->re) {
    data->re = pattern2re(data->pattern);
  }

  rc = pcre_exec(data->re, NULL, request->path_info, xmlStrlen(request->path_info), 0, 0, ovector, 30);
  if (rc > 0) {
    int i; xml_dtc_list *list; const char *result;
    list = xml_dtc_mappings_new_list(request->mappings);

    for (i = 0; i < rc; i++) {
      pcre_get_substring(request->path_info, ovector, rc, i, &result);
      xml_dtc_list_push_cast(list, (char *)result);
      free((char *)result);
    }
    i = xml_dtc_request_apply(request, data->sub_pipeline);
    xml_dtc_mappings_destroy_last(request->mappings);
    return i;
  } else
    return 0;
}

static int xml_dtc_match_wildcard_header(xml_dtc_caller *caller, xml_dtc_request *request) {
  int rc;
  int ovector[30];
  char *header;
  xml_dtc_matcher_caller_data *data = caller->_private;
  xml_dtc_component *matcher;

  matcher = (xml_dtc_component *) caller->component.object;
  if (!matcher->_private)
    return -1;
  header = xml_dtc_request_get_header(request, matcher->_private);
  if (! header)
    return -1;

  if (!data->re) {
    data->re = pattern2re(data->pattern);
  }

  rc = pcre_exec(data->re, NULL, header, xmlStrlen(header), 0, 0, ovector, 30);

  if (rc > 0) {
    int i; xml_dtc_list *list; const char *result;
    list = xml_dtc_mappings_new_list(request->mappings);

    for (i = 0; i < rc; i++) {
      pcre_get_substring(header, ovector, rc, i, &result);
      xml_dtc_list_push_cast(list, (char *)result);
      free((char *)result);
    }

    i = xml_dtc_request_apply(request, data->sub_pipeline);
    xml_dtc_mappings_destroy_last(request->mappings);
    return i;
  } else
    return 0;
}

static int xml_dtc_match_header(xml_dtc_caller *caller, xml_dtc_request *request) {
  int i;
  char *header;
  xml_dtc_list *list;
  xml_dtc_matcher_caller_data *data = caller->_private;
  
  header = xml_dtc_request_get_header(request, data->pattern);
  if (! header)
    return -1;

  list = xml_dtc_mappings_new_list(request->mappings);

  xml_dtc_list_push_cast(list, header);
  xml_dtc_list_push_cast(list, header);

  i = xml_dtc_request_apply(request, data->sub_pipeline);
  xml_dtc_mappings_destroy_last(request->mappings);
  return i;
}

static int xml_dtc_match_cookie(xml_dtc_caller *caller, xml_dtc_request *request) {
  int i;
  char *cookie;
  xml_dtc_list *list;
  xml_dtc_matcher_caller_data *data = caller->_private;

  cookie = xml_dtc_request_get_cookie(request, data->pattern);
  if (! cookie)
    return -1;

  list = xml_dtc_mappings_new_list(request->mappings);

  xml_dtc_list_push_cast(list, cookie);
  xml_dtc_list_push_cast(list, cookie);

  i = xml_dtc_request_apply(request, data->sub_pipeline);
  xml_dtc_mappings_destroy_last(request->mappings);
  return i;
}

static int xml_dtc_match_parameter(xml_dtc_caller *caller, xml_dtc_request *request) {
  int rc;
  int ovector[30];
  char *parameter;
  xml_dtc_component *matcher;
  xml_dtc_matcher_caller_data *data = caller->_private;

  matcher = (xml_dtc_component *) caller->component.object;
  if (!matcher->_private)
    return -1;
  parameter = xml_dtc_request_get_parameter(request, matcher->_private);
  if (! parameter)
    return -1;

  if (!data->re) {
    data->re = pattern2re(data->pattern);
  }

  rc = pcre_exec(data->re, NULL, parameter, xmlStrlen(parameter), 0, 0, ovector, 30);

  if (rc > 0) {
    int i; xml_dtc_list *list; const char *result;
    list = xml_dtc_mappings_new_list(request->mappings);

    for (i = 0; i < rc; i++) {
      pcre_get_substring(parameter, ovector, rc, i, &result);
      xml_dtc_list_push_cast(list, (char *)result);
      free((char *)result);
    }

    i = xml_dtc_request_apply(request, data->sub_pipeline);
    xml_dtc_mappings_destroy_last(request->mappings);
    return i;
  } else
    return 0;
}

static int xml_dtc_match_request_parameter(xml_dtc_caller *caller, xml_dtc_request *request) {
  xml_dtc_list *list;
  char *param;
  xml_dtc_matcher_caller_data *data = caller->_private;

  param = xml_dtc_request_get_parameter(request, data->pattern);
  if (param) {
    int i;
    list = xml_dtc_mappings_new_list(request->mappings);
    xml_dtc_list_push_cast(list, param);
    xml_dtc_list_push_cast(list, param);
    i = xml_dtc_request_apply(request, data->sub_pipeline);
    xml_dtc_mappings_destroy_last(request->mappings);
    return i;
  } else
    return 0;
}

static int xml_dtc_matchers_module_register(void) {
  return 0;
}

static const xml_dtc_caller_def xml_dtc_matchers_callers[] = {
  { "match", "matcher", (xml_dtc_handler_callback) xml_dtc_handler_match_callback },
  { NULL }
};

static const xml_dtc_component_def xml_dtc_matchers[] = {
  { "wildcard_uri", "matcher", xml_dtc_match, (xml_dtc_handler_callback) xml_dtc_handler_matcher_callback },
  { "wildcard_header", "matcher", xml_dtc_match_wildcard_header, (xml_dtc_handler_callback) xml_dtc_handler_matcher_wildcard_header_callback },
  { "request_parameter", "matcher", xml_dtc_match_request_parameter, (xml_dtc_handler_callback) xml_dtc_handler_matcher_callback },
  { "wildcard_request_parameter", "matcher", xml_dtc_match_parameter, (xml_dtc_handler_callback) xml_dtc_handler_matcher_wildcard_request_parameter_callback },
  { "header", "matcher", xml_dtc_match_header, (xml_dtc_handler_callback) xml_dtc_handler_matcher_callback },
  { "cookie", "matcher", xml_dtc_match_cookie, (xml_dtc_handler_callback) xml_dtc_handler_matcher_callback },
  { NULL }
};

xml_dtc_module xml_dtc_matchers_module = {
  XML_DTC_MODULE_STANDARD_STUFF,
  xml_dtc_matchers_module_register,
  NULL,
  xml_dtc_matchers_callers,
  xml_dtc_matchers
};
