/*
 * xml-dtc provides an XML-DTC implementation as a standalone program
 * It can generate a whole site tree from a sitemap
 * Copyright (C) 2003-2004 Mike Hommey <mh@glandium.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#define _GNU_SOURCE
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <xml-dtc.h>
#include <hash.h>
#include <components.h>

#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include <config.h>

xml_dtc_hash *queue = NULL;

char *uri2file(char *uri) {
  char *file = strdup(uri);
  char *query = index(file, '?');
  if (query)
    *query = '@';

  return file;
}

int create_dir(char *dir) {
  if (mkdir(dir, 0777) == -1) {
    if (errno == ENOENT) {
      char *dir2;
      int result;
      dir2 = rindex(dir, '/');
      dir2[0] = 0;
      result = create_dir(dir);
      dir2[0] = '/';
      return result;
    }
    return -1;
  }
  return 0;
}

int magic_open(char *filename) {
  int fd;
  char *file_only;
  
  file_only = rindex(filename, '/');
  if (file_only == NULL)
    file_only = filename;
  else {
    file_only[0] = 0;
    create_dir(filename);
    file_only[0] = '/';
    file_only++;
  }
  if (strlen(file_only) == 0) {
    char *new_file = malloc(strlen(filename) + 11);
    sprintf(new_file,"%sindex.html", filename);
    filename = new_file;
  }
  fd = open(filename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
  if (strlen(file_only) == 0) {
    free(filename);
  }
  return fd;
}

int test(xml_dtc_caller *caller, xml_dtc_request *request) {
  if (strcmp("serializer", caller->name) == 0) {
    if (request->xmltree) {
      xmlXPathContextPtr context;
      xmlXPathObjectPtr result;
      int i;
      xmlChar *string;

      context = xmlXPathNewContext(request->xmltree);
      xmlXPathRegisterNs(context, "xhtml", "http://www.w3.org/1999/xhtml");
      result = xmlXPathEvalExpression("//xhtml:a/@href|//xhtml:img/@src|//xhtml:link/@href", context);
      if (!xmlXPathNodeSetIsEmpty(result->nodesetval))
	for (i = 0; i < result->nodesetval->nodeNr; i++) {
	  string = xmlNodeListGetString(request->xmltree, result->nodesetval->nodeTab[i]->xmlChildrenNode, 1);
	  if (strstr(string, "://") == NULL) {
	    char *new_string;
	    char *offset, *old_offset;
	    int end = 0;

	    new_string = malloc(strlen(string) + strlen(request->path_info) + 2);
	    memset(new_string, 0, strlen(string) + strlen(request->path_info) + 2);
	    offset = rindex(request->path_info, '/');
	    if (offset == NULL)
	      offset = request->path_info;
	    if (strlen(request->path_info) != 0) {
	      memcpy(new_string, request->path_info, (int)offset - (int)request->path_info + 1);
	    }
	    old_offset = string;
	    while ((offset = index(old_offset, '/')) || (end == 0)) {
	      if (offset == 0) {
		offset = &old_offset[strlen(old_offset) - 1];
		end = 1;
	      }
	      if (strncmp(old_offset, "..", (int)offset - (int)old_offset) != 0) {
		memcpy(&new_string[strlen(new_string)], old_offset, (int)offset - (int)old_offset + 1);
	      } else {
		char *slash;
		if (strlen(new_string))
		  new_string[strlen(new_string) - 1] = 0;
		if ((slash = rindex(new_string, '/'))) {
		  memset(&slash[1], 0, strlen(new_string) - (int)slash + (int)new_string + 1);
		} else {
		  memset(new_string, 0, strlen(new_string));
		}
	      }
	      old_offset = &offset[1];
	    }
	    xml_dtc_hash_add_cast(queue, new_string, 1);
	    free(new_string);
	  }
	  xmlFree(string);
	}
      xmlXPathFreeObject(result);
      xmlXPathFreeContext(context);
    }
  }

  return 0;
}

void usage() {
  printf("Usage: xml-dtc [OPTIONS] sitemap [uri...]\n");
  printf("Valid options:\n");
  printf("  -h, --help       Shows this help message.\n");
  printf("  -v, --version    Shows version information.\n");
  printf("  -r, --recursive  Follow relative links in generated HTML pages.\n");
  printf("  -o [DIR]         Output generated files to specified directory.\n");
  printf("\n");
  printf("  sitemap          Sitemap file to use.\n");
  printf("  uri...           List of URIs used as start points in generation.\n");
  printf("\n");
  printf("Please report bugs at <%s>.\n", PACKAGE_BUGREPORT);
}

void version() {
  printf(PACKAGE_NAME " version " PACKAGE_VERSION "\n");
  printf("XML Data Chains Transformations.\n\n");
  printf("Please report bugs at <%s>.\n", PACKAGE_BUGREPORT);
}

int register_headers(xml_dtc_request *request) {
  xml_dtc_request_set_header(request, "user-agent", PACKAGE_NAME " " PACKAGE_VERSION);
  return 0;
}

int main(int argc, char *argv[]) {
  xml_dtc_sitemap *sitemap;
  xml_dtc_request *request;
  char *sitemap_file = NULL;
  char *output_dir = NULL;
  double t1, t2;
  struct timeval tp;
  int i = 0, e = 0;
  xml_dtc_resource *resource;

  resource = xml_dtc_resource_new();
  xml_dtc_resource_set_msg_verbosity(resource, XML_DTC_MSG_NOTICE);
  xml_dtc_init(resource);

  queue = xml_dtc_value_hash_new(xml_dtc.resource);

  while (++i < argc) {
    if (argv[i][0] == '-') {
      if ((strncmp(argv[i],"-o",3) == 0) || (strncmp(argv[i], "--output", 9) == 0)) {
	if (++i == argc) {
	  printf("Missing argument to option %s\n\n",argv[i-1]);
	  usage();
	  exit(1);
	}
	output_dir = argv[i];
      } else if ((strncmp(argv[i],"-r",3) == 0) || (strncmp(argv[i], "--recursive", 12) == 0)) {
	xml_dtc_register_hook(test);
      } else if ((strncmp(argv[i],"-h",3) == 0) || (strncmp(argv[i], "--help", 7) == 0)) {
	usage();
	exit(0);
      } else if ((strncmp(argv[i],"-v",3) == 0) || (strncmp(argv[i], "--version", 10) == 0)) {
	version();
	exit(0);
      } else {
	printf("Unknown option: %s\n\n", argv[i]);
	usage();
	exit(1);
      }
    } else {
      if (sitemap_file == NULL)
	sitemap_file = argv[i];
      else
	xml_dtc_hash_add_cast(queue, argv[i], 1);
    }
  }

  if (sitemap_file == NULL) {
    printf("Missing sitemap file name\n\n");
    usage();
    exit(1);
  }

  if (queue->num == 0)
    xml_dtc_hash_add_cast(queue, "", 1);
  
  if (output_dir)
    xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_NOTICE, "Output will be written in: %s.", output_dir);
  else
    xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_NOTICE, "Output will not be written.");

  xml_dtc_init_modules();

  sitemap = xml_dtc_sitemap_new_from_file(xml_dtc.resource, sitemap_file);

  if (!sitemap) {
    xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_ERROR, "Error while loading sitemap file.");
    return 1;
  }

  gettimeofday(&tp, '\0');
  t1 = tp.tv_sec + tp.tv_usec / 1000000.00;

  { int j;
  for (j = 0; j < 1; j++) {
    char *file, *path_info, *tmp;
    for(i = 0; i < queue->num; i++) {
      int fd;
      char *t;
      path_info = queue->list[i].key;
      if ((tmp = index(path_info, '?'))) {
	t = strndup(path_info, (int) tmp - (int) path_info);
      } else
	t = strdup(path_info);
      request = xml_dtc_request_new(xml_dtc.resource, "GET", t, t);
      free(t);
      if (tmp)
	xml_dtc_request_set_parameters_from_query_string(request, tmp + 1);
      request->register_headers = register_headers;
      xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_NOTICE, "Requesting: %s", path_info);
      xml_dtc_sitemap_request(sitemap, request);
      if (request->error.type != NULL) {
	printf("Error while requesting %s: %s\n", path_info, request->error.message);
	e++;
      } else if (output_dir) {
	tmp = uri2file(path_info);
	file = malloc(strlen(tmp) + strlen(output_dir) + 2);
	sprintf(file,"%s/%s", output_dir, tmp);
	printf("Writing %s\n", file);
	fd = magic_open(file);
	if (fd) {
	  if (request->content_length > 0)
	    write(fd, request->result, request->content_length);
	  else
	    printf("Can't determine size of output\n");
	  close(fd);
	}
	free(tmp);
	free(file);
      }
      xml_dtc_request_free(request);
    }
  }
  }
  gettimeofday(&tp, '\0');
  t2 = tp.tv_sec + tp.tv_usec / 1000000.00;

  fprintf(stderr,"%d x xml_dtc_run(xml_dtc, sitemap, request): %.10f (%.0f/s) (%d error%c)\n", i, t2 - t1, (double)(i / (t2 - t1)), e, (e > 1) ? 's' : 0);

  xml_dtc_sitemap_free(sitemap);

  xml_dtc_hash_free(queue);

#ifdef OBJECT_DEBUG
  xml_dtc_objects_dump();
#endif

  xml_dtc_cleanup();

  xmlCleanupParser();

  xmlMemoryDump();

  return 0;
}
